
%On the semantics of 'case' pattern matching

% on pattern matching with multiple occurrences of the same variable in pattern:
local proc {TestException A B}
	 local R = r(A B) in
	    case R of
	       r(X X) then {Browse X}
	    else {Browse 'no match'}
	    end
	 end
      end
in {TestException a a}
   {TestException a b}
end


% according to simple 'case' semantics, the code would be equivalent to:
local proc {TestException A B}
	 local X=A X=B in {Browse X} end
      end
in {TestException a a}
   {TestException a b} %throws "a=b" failure (incompatible assignment)
end

% according to more complex 'case' semantics, it is equivalent to:
local proc {TestException A B}
	 try  
	    local X=A X=B in {Browse X} end
	 catch E then
	    {Browse 'no match'}
	 end
      end
in {TestException a a}
   {TestException a b}
end

% however, courtesy of T.J. Lipscomb, if the 'case' code throws an
% exception after successfully pattern matching:
local proc {TestException A B}
	 local R = r(A B) in
	    case R of
	       r(X X) then {Browse X}
	                   raise e end
	    else {Browse 'no match'}
	    end
	 end
      end
in try {TestException a a}
   catch E then {Browse E}
   end
   try {TestException a b}
   catch E then {Browse E}
   end
end

% in the 'complex' translation, the 'else' code would be "erroneously"
% executed, e.g.:
local proc {TestException A B}
	 try
	    local X=A X=B in
	       {Browse X}
	       raise e end
	    end
	 catch E then
	    {Browse 'no match'}
	 end
      end
in try {TestException a a}
   catch E then {Browse E}
   end
   try {TestException a b}
   catch E then {Browse E}
   end
end

% we could "re-raise" the exception in the complex translation, but both
% "exception handling" codes would be executed, which is incorrect:
local proc {TestException A B}
	 try
	    local X=A X=B in
	       {Browse X}
	       raise e end
	    end
	 catch E then
	    {Browse 'no match'}
	    raise E end
	 end
      end
in try {TestException a a}
   catch E then {Browse E}
   end
   try {TestException a b}
   catch E then {Browse E}
   end
end

% so, here is a potential solution; we need to distinguish between user
% exceptions and non-user exceptions, and execute the 'else' clause
% only on non-user exceptions:
local proc {TestException A B}
	 try
	    local X=A X=B in
	       {Browse X}
	       raise user(e) end
	    end
	 catch user(E) then
	    raise E end
	 [] E then
	    {Browse 'no match'}
	 end
      end
in try {TestException a a}
   catch E then {Browse E}
   end
   try {TestException a b}
   catch E then {Browse E}
   end
end

% Dylan's example however breaks this translation approach:
local proc {TestException A B}
	 local R = r(A B) in
	    case R of
	       r(X X) then {Browse X}
	                   X = 2
	    else {Browse 'no match'}
	    end
	 end
      end
in try {TestException a a}
   catch E then {Browse E}
   end
   try {TestException a b}
   catch E then {Browse E}
   end
end

% problem is user cannot catch system exception:
local proc {TestException A B}
	 try
	    local X=A X=B in 
	       {Browse X}
	       X = 2  % produces a system exception...
	       raise user(e) end
	    end
	 catch user(E) then
	    raise E end
	 [] E then
	    {Browse 'no match'}
	 end
      end
in try {TestException a a}
   catch E then {Browse E}  % ...that should be caught here.
   end
   try {TestException a b}
   catch E then {Browse E}
   end
end

% here would be a solution (assuming we could catch system failures):
local proc {TestException A B}
	 try
	    local X=A X=B in 
	       try
		  {Browse X}
		  X = 2  % produces a system exception...
		  raise user(e) end
	       catch failure(E) then
		  raise user(failure(E)) end
	       end
	    end
	 catch user(E) then
	    raise E end
	 [] E then
	    {Browse 'no match'}
	 end
      end
in try {TestException a a}
   catch E then {Browse E}  % ...that should be caught here.
   end
   try {TestException a b}
   catch E then {Browse E}
   end
end

%However, while this works:
declare
fun {One} 1 end
fun {Two} 2 end
try {One} = {Two}
catch E then {Browse E}
end

%This does not(!):
declare
fun {One} 1 end
fun {Two} 2 end
try {One} = {Two}
catch failure(E) then {Browse E}
end

%And this does not either(!):
declare
fun {One} 1 end
fun {Two} 2 end
try {One} = {Two}
catch E then
   case E of failure(F) then {Browse F}  % then clause should execute
   else {Browse E} % but else clause does ;-(
   end
end

% on pattern matching with multiple occurrences of the same variable in pattern REVISITED:
local proc {TestException A B}
	 local R = r(A B) in
	    case R of
	       r(X X) then {Browse X}
	    else {Browse 'no match'}
	    end
	 end
      end
in {TestException a a}
   {TestException a b}
end

% according to CTM Chapter 2 'case' semantics, since Label and Arity correspond,
% the code would be (wrongly) equivalent to:
local proc {TestException A B}
	 local X=A X=B in {Browse X} end
      end
in {TestException a a}
   {TestException a b} %throws "a=b" failure (incompatible assignment)
end

% According to unification semantics (see Slide 4 in Lecture Slides),
% {TestException A B} would be translated to:
local proc {TestException A B}
	 try local r(X X)=r(A B) in {Browse X} end
	 catch E then {Browse 'no match'} end
      end
in {TestException a a}
   {TestException a b}
end

% if you throw an exception in the case code,
local proc {TestException A B}
	 local R = r(A B) in
	    case R of r(X X) then
	       {Browse X}
	       raise e end   % user exception
	    else {Browse 'no match'}
	    end
	 end
      end
in  try {TestException a a}
   catch E then {Browse E}
   end
   try {TestException a b}
   catch E then {Browse E}
   end
end

%you generate code:
local proc {TestException A B}
	 try local r(X X)=r(A B) in
		try
		   {Browse X}
		   raise e end
		catch E then
		   raise user(E) end
		end
	     end
	 catch user(E) then raise E end
	 [] E then {Browse 'no match'}
	 end
      end
in try {TestException a a}
   catch E then {Browse E}
   end
   try {TestException a b}
   catch E then {Browse E}
   end
end

% which also works with "system" exceptions:
local proc {TestException A B}
	 local R = r(A B) in
	    case R of r(X X) then
	       {Browse X}
	       X = 2     % system exception on user code
	    else {Browse 'no match'}
	    end
	 end
      end
in  try {TestException a a}
   catch E then {Browse E}
   end
   try {TestException a b}
   catch E then {Browse E}
   end
end

%which would generate:
local proc {TestException A B}
	 try local r(X X)=r(A B) in  %  <l>(<f1>:<x1> ...<fn>:<xn>) = <x> unification
		try
		   {Browse X}      % <s1>
		   X=2             %
		catch E then
		   raise user(E) end
		end
	     end
	 catch user(E) then raise E end  % captures errors in user code and re-raises them
	 [] E then {Browse 'no match'}   % <s2>  only executes if error in unification
	 end
      end
in try {TestException a a}
   catch E then {Browse E}
   end
   try {TestException a b}
   catch E then {Browse E}
   end
end


% does pattern matching complete partial values?
declare X T = t(a X)
case T of
   t(Y Y) then {Browse X}  % No:  pattern matching suspends until it can (not) match.
   else {Browse 'no match'}
end

X = a  % this enables matching

X = b  % this does not enable matching

% it is possible to pattern match with partial values
declare X T = t(a X)
case T of
   t(a Z) then {Browse 'match with partial value'}  % Yes.
   else {Browse 'no match'}
end

{Browse X}

% so, it looks like our translation still would have trouble modeling the case statement
% because of not suspending in the following case:
declare X
local t(Y Y) = t(a X) in {Browse X} end